#ifndef HATPPreprocessor_HH
#define HATPPreprocessor_HH

#include <iostream>
#include <fstream>
#include <algorithm>
#include <string>
#include <utility>

#include <cstdio>

#include <boost/regex.hpp>

/** Class representing a range (or a block of lines) in the complete file and its corresponding range in the original file.
 * Stores a the range of lines in the complete file, the corresponding range of line in the original file
 * and the name of the original file.
 */
class Range{
private:
	/** First line (of this block) in the complete file */
	size_t beginComplete;
	/** Last line (of this block) in the complete file */
	size_t endComplete;
	
	/** First line (of this block) in the original file */
	size_t beginReal;
	/** Last line (of this block) in the original file */
	size_t endReal;
	/** Name of the original file */
	std::string realFilename;

	/** Default constructor: disabled. */
	Range();

public:
	/** Constructor: initialises every field of the Range class.
	 * @param beginComplete Line number of the block beginning in the complete file
	 * @param beginReal Line number of the block beginning in the original file
	 * @param realFilename Name of the original file
	 */
	Range(size_t beginComplete, size_t beginReal, const std::string& realFilename): beginComplete(beginComplete), endComplete(beginComplete), beginReal(beginReal), endReal(beginReal), realFilename(realFilename) {}

	/** Copy constructor.
	 * @param range Range to copy
	 */
	Range(const Range& range): beginComplete(range.beginComplete), endComplete(range.endComplete), beginReal(range.beginReal), endReal(range.endReal), realFilename(range.realFilename) {}

	/** Assignment operator, uses the copy-and-swap idiom.
	 * @param range Range to copy
	 * @return The new Range
	 */
	Range& operator=(Range range){
		std::swap(*this, range);
		return *this;
	}

	/** Update the last line number of both ranges (see Preprocessor::includeFile).
	 */
	void increaseLineNumber(){
		++endComplete;
		++endReal;
	}

	/** Return the last line number for the complete file.
	 * @return Line number
	 */
	size_t lastLine() const {
		return endComplete;
	}

	/** Tries to return the line number in the original file corresponding to the line number in the complete file given as parameter.
	 * @param line Line number in the complete file
	 * @return A pair with the filename and the orinigal line number if a match is found, otherwise return an empty pair <"", 0>
	 */
	std::pair<std::string, size_t> getLine(size_t line) const {
		if(line<beginComplete || line>endComplete){
			return std::pair<std::string, size_t>("", 0);
		}

		size_t offset=line-beginComplete;

		if(beginReal+offset>endReal){
			return std::pair<std::string, size_t>("", 0);
		}
		return std::make_pair(realFilename, beginReal+offset);
	}

	/** Test if the (complete file) line number given as input is in this Range.
	 * @param line Line number to test
	 * @return True if the number >= first line in the complete file AND number <= last line in the complete file
	 */
	bool operator==(size_t line) const {
		return(line>=beginComplete&&line<=endComplete);
	}
};

/** A basic (and inaccurate) preprocessor for HATP language.
 * The preprocessor can parse "include" clause. The principle is that for each include clause
 * the preprocessor add the corresponding file content in the current file. Two structures are
 * kept: the complete file (after all includes) and a structures that allows to know the origin
 * of a code line.
 * To understand the matching between a code line and its origin see Range.
 */
class Preprocessor{
private:
	/** Path to the main domain file, see constructor */
	std::string path;
	/** Store the file as it is expanded */
	std::ofstream complete;
	/** Store the resulting file after the includes have been expanded */
	std::ifstream completeFile;
	/** Path to the complete file */
	std::string completeDomainFilename;
	/** Associate to each line from the complete domain its origin (file and line) */
	std::vector<Range> content;
	/** Whether it was possible to parse the domain files or not */
	bool worked;

	/** Default constructor: disabled. */
	Preprocessor();
	/** Copy constructor: disabled.
	 * @param p Disabled
	 */
	Preprocessor(const Preprocessor& p);
	/** Assignment operator: disabled.
	 * @param p Disabled
	 * @return Disabled
	 */
	Preprocessor& operator=(const Preprocessor& p);

public:
	/** Constructor: parse the input domain file, search for include clause and expand them.
	 * If everything work as expected, completeFile contains the input file stream that can be used by the main parser.
	 * It also set "worked" to true and creates the vector containing the match between compete file file and origin of the line.
	 * It generates a file (name in completeDomainFilename) to store the expanded file, the input file stream points to this file.
	 * @param domainFile Complete path to a domain file to parse
	 */
	Preprocessor(const std::string& domainFile): completeDomainFilename(domainFile+".complete") {
		complete.open(completeDomainFilename.c_str(), std::ios_base::in | std::ios_base::out | std::ios_base::trunc);

		if(!complete.is_open()){
			std::cerr<<"Error while preparing the temporary file : "<<completeDomainFilename<<std::endl;
			worked=false;
		}else{
			//Extract the path to the file (all the text before the last '/')
			size_t length=domainFile.rfind("/");
			if(length!=0){
				path=domainFile.substr(0, length+1);
			}

			//Expand all the include clauses
			worked=includeFile(domainFile);

			complete.close();

			//If everything worked as expected, open the complete file and create the completeFile input stream so the main parser can use it
			if(worked){
				completeFile.open(completeDomainFilename.c_str());
				if(!completeFile.is_open()){
					worked=false;
				}
			}
		}
	}

	/** Destrcutor: Remoce the generated file and close the ifstream. */
	~Preprocessor(){
		//Delete the complete file generated
		remove(completeDomainFilename.c_str());
		completeFile.close();
	}

	/** Tell whether the pre-proccessing phase happened as expected.
	 * @return Worked, the variable storing the result
	 */
	bool isReady(){
		return worked;
	}

private:
	/** Expand all the include clauses.
	 * The algorithm is recursive: it is called everytime an include clause is found.
	 * See Preprocessor(const std::string& domainFile)
	 * @param includeName Domain file to expand
	 * @return True if the process worked as expexted, false otherwise
	 */
	bool includeFile(const std::string& includeName){
		std::ifstream include;
		include.open(includeName.c_str());

		if(!include.is_open()){
			std::cerr<<"Problem while openning the file \""<<includeName<<"\""<<std::endl;
			return false;
		}

		size_t lineNumber=1; //Line number in the current file
		bool fileLinesStored=false; //The line number have been stored for this file for this block (between includes lines or the begining/end)
		bool newIncludedFile=false; //A file has just been included
		std::string line; //Holds the current extracted line (to paste it if necessary)
		size_t it;
		while(!include.eof()){
			//Algo works on one line at a time
			std::getline(include, line);
			if(include.eof()){ continue; }
			it=line.find("#include");

			//If the line contains the text "#include": expand it
			if(it!=std::string::npos){
				boost::cmatch what;
				
				//If a correct include clause is found
				if(boost::regex_match(line.c_str(), what, boost::regex("[[:space:]]*[[.number-sign.]]include[[:space:]]\".*\"[[:space:]]*"))){
					size_t first=line.find("\"")+1;
					size_t last=line.rfind("\"");
					size_t length=last-first;
					if(length!=0){
						std::string includeFilename=line.substr(first, length);
						//Recursive call to include the file
						if(!includeFile(path+includeFilename)){
							std::cerr<<includeName<<":"<<lineNumber<<" Error : Impossible to include \""<<includeFilename<<"\""<<std::endl;
							return false;
						}
						newIncludedFile=true;
					}else{
						std::cerr<<includeName<<":"<<lineNumber<<" Error : Wrong included file"<<std::endl;
						return false;
					}
				}else{
					std::cerr<<includeName<<":"<<lineNumber<<" Error : Include has wrong format, should be #include \".....\""<<std::endl;
					return false;
				}

			//If the line does not contain an include clause:
			//copy the line as is to the complete file and keep track of its origin
			}else{
				complete<<line<<std::endl;
				//If the line numbers for this file have not been stored, create a line in the content vector
				if(!fileLinesStored){
					if(content.empty()){
						content.push_back(Range(1, lineNumber, includeName)); //First range to be inserted
					}else{
						size_t previous=content.back().lastLine()+1;
						content.push_back(Range(previous, lineNumber, includeName)); //Other ranges
					}
					fileLinesStored=true;
					newIncludedFile=false;
				}else{
					//If a new file has been inserted create a new range for this file after the included file
					if(newIncludedFile){
						size_t previous=content.back().lastLine()+1;
						content.push_back(Range(previous, lineNumber, includeName));
						fileLinesStored=true;
						newIncludedFile=false;
					}else{
						//Update the last range line number if a line has been copied
						content.back().increaseLineNumber();
					}
				}
			}
			++lineNumber;
		}
		include.close();

		return true;
	}

public:
	/** Get an input file stream to the complete domain file.
	 * @return An ifstream to the complete file
	 */
	std::ifstream& getCompleteDomain(){
		return completeFile;
	}

	/** Get the original file and line number for a line number in the complete file.
	 * @param line Line number in thte complete file
	 * @return A pair with the filename and the line number
	 */
	 std::pair<std::string, size_t> getLine(size_t line) const {
		std::vector<Range>::const_iterator it;
		it=std::find(content.begin(), content.end(), line);
		if(it!=content.end()){
			return it->getLine(line);
		}
		return std::pair<std::string, size_t>("", 0);
	}
};

#endif //HATPPreprocessor_HH
