/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef HATP_CHOICE_POINT_HH_
#define HATP_CHOICE_POINT_HH_

#include<vector>
#include<string>

#include "LightWorldBase.hh"
#include "Task.hh"
#include "PartialPlanAppraiser.hh"
#include "TaskListNode.hh"
#include "TaskList.hh"
#include "PlanTree.hh"

/** Represent a choice the planner faced during the planning phase, allows to backtrack.
 * Each choice becomes a node in a tree. So a node has a parent and a set of children.
 * A node represent the state of the planner when the choice occured, so it stores the state (world base, current plan, ...).
 */
class ChoicePoint
{
public:
	/** Type of the choice point.
	 */
	enum Type {ROOT_TYPE, ///< The root of ChoiceReminder
		DECOMPO_TYPE, ///< Planning faced a decomposition choice
		PARALLEL_TYPE, ///< Planning faced a parallel choice
		GEOMETRIC_TYPE ///< A task was projected on the geometry
	};

	/** Possible states of the choice point.
	 */
	enum Tag {CURRENT_TAG, ///< Planning's current point
		UNEXPLORED_TAG, ///< Choice not explored
		EXPLORED_TAG, ///< Choice expolored
		SUCCESS_TAG, ///< Choice gave a possible plan
		FAILED_TAG, ///< Choice did not allow to reach a plan
		COST_EXCEEDED_TAG ///< Choice current cost is higher than best plan cost, do not explore further
	};
private:
	/** The type of the choice point.
	 */
	Type type;

	/** Choice point state.
	 */
	Tag tag;

	/** The WorldBase associated to the choice point.
	 */
	LightWorldBase* worldbase;

	/** Tthe TaskList associated to the choice point.
	 */
	TaskList* tasklist;

	/** The vector defining order in which tasks have been projected.
	 */
	std::vector<unsigned long int>* projection_order;

	/** Id of the task (member of tasklist) associated to the choice point.
	 */
	unsigned long int task;

	/** The PlanTree associated to the choice point.
	 */
	PlanTree* tree;

	/** The cost of the projection.
	 */
	PartialPlanAppraiser* partial_eval;

	/** The children of the choice point.
	 */
	std::vector<ChoicePoint*> next;

	/** The father of the choice point.
	 */
	ChoicePoint* father;

	double cost;

public:
	explicit ChoicePoint(ChoicePoint::Type sType, LightWorldBase const& wb, TaskList const& TL,
		std::vector<unsigned long int> const& po, PlanTree const& PT, unsigned long int TaskID,
		PartialPlanAppraiser const& pEval);
	explicit ChoicePoint(ChoicePoint::Type sType, LightWorldBase const& wb, TaskList* TL,
		std::vector<unsigned long int> const& po, PlanTree* PT, unsigned long int TaskID,
	PartialPlanAppraiser const& pEval);
	explicit ChoicePoint(ChoicePoint const& cp);

	Tag const& getTag() const;
	bool isExplored() const;
	void setTag(Tag sTag);
	Tag propagateTags();

	void addChild(ChoicePoint* CP);
	void setFather(ChoicePoint* CP);
	ChoicePoint* getFather() const;
	std::vector<ChoicePoint*> const& getChildren() const;
	void deleteChildren();
	bool hasNoCurrentChild() const;

	ChoicePoint::Type getType() const;
	LightWorldBase* getWorldBase();
	LightWorldBase const* getWorldBase() const;
	TaskList* getTaskList();
	TaskList const* getTaskList() const;
	std::vector<unsigned long int> * getProjectionOrder() const;
	unsigned long int getTask() const;
	PlanTree* getTree();
	PlanTree const* getTree() const;
	PartialPlanAppraiser const* getPartialEval() const;
	PartialPlanAppraiser* getPartialEval();

	std::string toString() const;
	static std::string typeToString(Type type);
	static std::string tagToString(Tag tag);
	~ChoicePoint();

	double getCost() const;
	void setCost(double cost);
};

/** Return the tag.
 * @return tag
 */
inline ChoicePoint::Tag const& ChoicePoint::getTag() const
{
	return tag;
}

/** Return the tag.
 * @return tag
 */
inline bool ChoicePoint::isExplored() const
{
	return (tag==FAILED_TAG ||  tag==SUCCESS_TAG ||  tag==EXPLORED_TAG ||  tag==COST_EXCEEDED_TAG);
}

/** Return the type.
 * @return type
 */
inline ChoicePoint::Type ChoicePoint::getType() const
{
	return type;
}

/** Set the tag to the given one.
 * @param sTag The new tag
 */
inline void ChoicePoint::setTag(ChoicePoint::Tag sTag)
{
	tag = sTag;
	if(sTag!=CURRENT_TAG && sTag!=UNEXPLORED_TAG){
		cost=-1;
	}
}

/** Add a child to this choice point.
 * Adds CP in next, sets CP->father to this
 * @param CP The child point
 */
inline void ChoicePoint::addChild(ChoicePoint* CP)
{
	next.push_back(CP);
	CP->setFather(this);
}

/** Set this choice point father.
 * @param CP The father point
 */
inline void ChoicePoint::setFather(ChoicePoint* CP)
{
	father = CP;
}

/** Return the father.
 * @return The father
 */
inline ChoicePoint* ChoicePoint::getFather() const
{
	return father;
}

/** Return the list of next choice points (children of this node).
 * @return Next choice points
 */
inline std::vector<ChoicePoint*> const& ChoicePoint::getChildren() const
{
	return next;
}

/** Return the LightWorldBase.
 * @return The LightWorldBase
 */
inline LightWorldBase* ChoicePoint::getWorldBase()
{
	return worldbase;
}

/** Return the LightWorldBase (const version).
 * @return The LightWorldBase
 */
inline LightWorldBase const* ChoicePoint::getWorldBase() const
{
	return worldbase;
}

/** Return the TaskList.
 * @return The TaskList
 */
inline TaskList* ChoicePoint::getTaskList()
{
	return tasklist;
}

/** Return the TaskList (const version).
 * @return The TaskList
 */
inline TaskList const* ChoicePoint::getTaskList() const
{
	return tasklist;
}

/** Return the projection order.
 * @return The projection order
 */
inline std::vector<unsigned long int> * ChoicePoint::getProjectionOrder() const
{
	return projection_order;
}

/** Return the associated task.
 * @return The task
 */
inline unsigned long int ChoicePoint::getTask() const
{
	return task;
}

/** Return the PlanTree.
 * @return The PlanTree
 */
inline PlanTree* ChoicePoint::getTree()
{
	return tree;
}

/** Return the PlanTree (const version).
 * @return The PlanTree
 */
inline PlanTree const* ChoicePoint::getTree() const
{
	return tree;
}

/** Return the PartialPlanAppraiser.
 * @return The PartialPlanAppraiser
 */
inline PartialPlanAppraiser* ChoicePoint::getPartialEval()
{
	return partial_eval;
}

/** Return the PartialPlanAppraiser (const version).
 * @return The PartialPlanAppraiser
 */
inline PartialPlanAppraiser const* ChoicePoint::getPartialEval() const
{
	return partial_eval;
}

/** Destructor.
 * Deletes WorldBase, TaskList, PlanTree and PartialPlanAppraiser
 */
inline ChoicePoint::~ChoicePoint()
{
	delete worldbase;
	delete tasklist;
	delete tree;
	delete partial_eval;
}

inline double ChoicePoint::getCost() const{
	return cost;
}

inline void ChoicePoint::setCost(double new_cost){
	cost=new_cost;
}

#endif
