/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef HATP_CHOICE_REMINDER_HH_
#define HATP_CHOICE_REMINDER_HH_

#include<vector>
#include<set>
#include<map>

#include "ChoicePoint.hh"
#include "WorldBase.hh"
#include "LightWorldBase.hh"
#include "PartialPlanAppraiser.hh"
#include "TaskList.hh"
#include "PlanTree.hh"

/** This class stores all the choice points created by the planner.
 * The ChoicePoints are stored in a tree fashion
 * There are many getters to lots of the current_point attributes
 */
class ChoiceReminder
{
private:
	/** The root node of the ChoiceReminder.
	*/
	ChoicePoint* root_point;

	/** The current node of the ChoiceReminder.
	*/
	ChoicePoint* current_point;

	/** The list of all ChoicePoints.
	*/
	std::set<ChoicePoint*> points;
	
	std::string recursiveBranchPrint(ChoicePoint* point, std::string offset) const;

public:
	explicit ChoiceReminder(WorldBase const& wb, TaskList const& TL, PartialPlanAppraiser const& ppa);
	ChoiceReminder(ChoiceReminder const& cr);
	
	void addChoicesToCurrentNode(std::vector<ChoicePoint*> const& vec);
	
	bool backtrack();
	bool update();
	ChoicePoint* getRootPoint() const;
	ChoicePoint* getCurrentPoint() const;
	std::set<ChoicePoint*> const& getChoicePoints() const;
	unsigned long int getNbChoicePoints() const;
	ChoicePoint::Type getCurrentPointType();

	void setCurrentPointTag(ChoicePoint::Tag tag);
	ChoicePoint::Tag getCurrentPointTag();
	std::vector<unsigned long int> getCurrentFatherChoices() const;
	bool selectUnexploredChildOfCurrentPoint();

	LightWorldBase* getCurrentWorldBase();
	LightWorldBase const* getCurrentWorldBase() const;
	TaskList* getCurrentTaskList();
	TaskList const* getCurrentTaskList() const;
	std::vector<unsigned long int> * getCurrentProjectionOrder() const;
	unsigned long int getCurrentTask() const;
	PlanTree* getCurrentPlanTree();
	PlanTree const* getCurrentPlanTree() const;
	PartialPlanAppraiser* getCurrentPartialEval();
	PartialPlanAppraiser const* getCurrentPartialEval() const;

	std::string toString() const;

	~ChoiceReminder();
};

/** Return the root node.
 * @return Root ChoicePoint
 */
inline ChoicePoint* ChoiceReminder::getRootPoint() const
{
	return root_point;
}

/** Return the current node.
 * @return Current ChoicePoint
 */
inline ChoicePoint* ChoiceReminder::getCurrentPoint() const
{
	return current_point;
}

/** Return tag of the current point.
 * @return Current ChoicePoint::Tag
 */
inline ChoicePoint::Tag ChoiceReminder::getCurrentPointTag()
{
	return current_point->getTag();
}

/** Return type of the current point.
 * @return Current ChoicePoint::Type
 */
inline ChoicePoint::Type ChoiceReminder::getCurrentPointType()
{
	return current_point->getType();
}


/** Set the current node tag.
 * @param tag New tag for the ChoicePoint
 */
inline void ChoiceReminder::setCurrentPointTag(ChoicePoint::Tag tag)
{
	current_point->setTag(tag);
}

/** Return the LightWorldBase of the current point.
 * @return Current ChoicePoint world base
 */
inline LightWorldBase* ChoiceReminder::getCurrentWorldBase()
{
	return current_point->getWorldBase();
}

/** Return the LightWorldBase of the current point. (const version)
 * @return Current ChoicePoint world base
 */
inline LightWorldBase const* ChoiceReminder::getCurrentWorldBase() const
{
	return current_point->getWorldBase();
}

/** Return the TaskList of the current point.
 * @return Current ChoicePoint TaskList
 */
inline TaskList* ChoiceReminder::getCurrentTaskList()
{
	return current_point->getTaskList();
}

/** Return the LightWorldBase of the current point. (const version)
 * @return Current ChoicePoint world base
 */
inline TaskList const* ChoiceReminder::getCurrentTaskList() const
{
	return current_point->getTaskList();
}

/** Return the projection order of the current point.
 * @return Current ChoicePoint projection order
 */
inline std::vector<unsigned long int> * ChoiceReminder::getCurrentProjectionOrder() const
{
	return current_point->getProjectionOrder();
}

/** Return the Task ID associated to the current point.
 * @return Current ChoicePoint associated Task ID
 */
inline unsigned long int ChoiceReminder::getCurrentTask() const
{
	return current_point->getTask();
}

/** Return the PlanTree of the current point.
 * @return Current ChoicePoint PlanTree
 */
inline PlanTree* ChoiceReminder::getCurrentPlanTree()
{
	return current_point->getTree();
}

/** Return the PlanTree of the current point. (const version)
 * @return Current ChoicePoint PlanTree
 */
inline PlanTree const* ChoiceReminder::getCurrentPlanTree() const
{
	return current_point->getTree();
}

/** Return the PartialPlanAppraiser of the current point.
 * @return Current ChoicePoint PartialPlanAppraiser
 */
inline PartialPlanAppraiser* ChoiceReminder::getCurrentPartialEval()
{
	return current_point->getPartialEval();
}

/** Return the PartialPlanAppraiser of the current point. (const version)
 * @return Current ChoicePoint PartialPlanAppraiser
 */
inline PartialPlanAppraiser const* ChoiceReminder::getCurrentPartialEval() const
{
	return current_point->getPartialEval();
}

/** Return all the choice points in this reminder.
 * @return Set of ChoicePoint
 */
inline std::set<ChoicePoint*> const& ChoiceReminder::getChoicePoints() const
{
	return points;
}

/** Return the number of choice points in the choice reminder.
 * @return Number of ChoicePoint
 */
inline unsigned long int ChoiceReminder::getNbChoicePoints() const
{
	return points.size();
}

#endif
