#if !defined HATP_HTNBASE_HH_
#define HATP_HTNBASE_HH_

#include<string>
#include<map>

#include "Task.hh"
#include "BeliefMethod.hh"

/** The list of tasks in the HTN tree (domain representation). If necessary/used, it also contains the belief management method for the domain.
 */
class HTNBase {

private:
	/** Default task used for empty Task.
	 */
	Task defaultTask;

	/** List of all tasks in the tree.
	 */
	std::map<std::string, Task*> database;

	/** Belief management method.
	 */
	BeliefMethod* beliefMethod;

public:
	explicit HTNBase();
	HTNBase(HTNBase const& htnb);
	HTNBase& operator=(HTNBase const& htnb2copy);
	
	void add(Task* e);
	void setBeliefMethod(BeliefMethod* bm);
	
	bool existingTask(std::string const& name) const;
	Task& find(std::string const& name);
	Task const& findConst(std::string const& name);
	Task const& find(std::string const& name) const;
	
	std::map<std::string, Task*> const& getDatabase() const;
	BeliefMethod* getBeliefMethod() const;
	
	std::string toString() const;
	~HTNBase();
};

/** Default constructor.
 * Empty base (defautTask is automatically set to an empty Task)
 */
inline HTNBase::HTNBase() {}

/** Return database.
 * @return Set of tasks and belief management method
 */
inline std::map<std::string, Task*> const& HTNBase::getDatabase() const {
	return database;
}

/** Find a task using its name.
 * @param name Name of the task to find
 * @return The task or if not found returns defaultTask (and prints an error message)
 */
inline Task& HTNBase::find(std::string const& name) {
	std::map<std::string, Task*>::iterator it = database.find(name);
	if(it != database.end()) return *(it->second);
	std::cerr << "ERROR: trying to access to unknown task \"" << name << "\"" << std::endl;
	std::cerr << "default task return instead" << std::endl;
	return defaultTask;
}

/** Find a task using its name (fully const version).
 * @param name Name of the task to find
 * @return The (const) task or if not found returns defaultTask (and prints an error message)
 */
inline Task const& HTNBase::find(std::string const& name) const {
	std::map<std::string, Task*>::const_iterator it = database.find(name);
	if(it != database.end()) return *(it->second);
	std::cerr << "ERROR: trying to access to unknown task \"" << name << "\"" << std::endl;
	std::cerr << "default task return instead" << std::endl;
	return defaultTask;
}

/** Find a task using its name (const version).
 * @param name Name of the task to find
 * @return The (const) task or if not found returns defaultTask (and prints an error message)
 */
inline Task const& HTNBase::findConst(std::string const& name) {
	std::map<std::string, Task*>::const_iterator it = database.find(name);
	if(it != database.end()) return *(it->second);
	std::cerr << "ERROR: trying to access to unknown task \"" << name << "\"" << std::endl;
	std::cerr << "default task return instead" << std::endl;
	return defaultTask;
}

#endif
