/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>
#include "PlanStockManager.hh"

using namespace std;
using namespace boost::posix_time;

/** Add a plan to the stock.
 * Check if p is already in plans, if not add it.
 * If the plan has a score worst than the actual best one, do not keep it in memory (but increas numberOfPlans)
 * A message linked to start_time is created if the p is added.
 * @param p Plan to add
 */
void PlanStockManager::addPlan(Plan* p) {
	ostringstream lm;
	if(plans.empty()) { // we have found a first plan
		++NumberOfPlans;
		solutionExists = true;
		plans.insert(make_pair(p->getScore(), p));
		lm << "A first plan (id=" << p->getID() << ") has been found, added after ";
		lm << (microsec_clock::local_time()-start_time).total_milliseconds() << " ms ";
		std::map<std::string, float> const& scoreDetails = p->getScoreDetails();
		std::map<std::string, float>::const_iterator it;
		for(it = scoreDetails.begin() ; it != scoreDetails.end() ; ++it) lm<<"("<<it->first<<" = "<<it->second<<")";
		lm << "(score = " << p->getScore() << ")";
	}
	else {
		// we compute contribution
		float contrib = p->determineContribution(plans.begin()->second->getScoreDetails());
		if(contrib < 0.){ // p is worse that current plans, we free memory
			delete p;
			++NumberOfPlans;
		}
		else if(contrib == 0.) { // p is similar to plans currently in memory
			++NumberOfPlans;
			plans.insert(make_pair(p->getScore(), p));
			lm << "A similar plan (id=" << p->getID() << ") has been found, added after ";
			lm << (microsec_clock::local_time()-start_time).total_milliseconds() << " ms ";
			std::map<std::string, float> const& scoreDetails = p->getScoreDetails();
			std::map<std::string, float>::const_iterator it;
			for(it = scoreDetails.begin() ; it != scoreDetails.end() ; ++it) lm<<"("<<it->first<<" = "<<it->second<<")";
			lm << "(score = " << p->getScore() << ")";
		}
		else { // we have found a better plan, we clear plans and we add p in it
			++NumberOfPlans;
			this->clear();
			plans.insert(make_pair(p->getScore(), p));
			lm << "A better plan (id=" << p->getID() << ") has been found, added after ";
			lm << (microsec_clock::local_time()-start_time).total_milliseconds() << " ms ";
			std::map<std::string, float> const& scoreDetails = p->getScoreDetails();
			std::map<std::string, float>::const_iterator it;
			for(it = scoreDetails.begin() ; it != scoreDetails.end() ; ++it) lm<<"("<<it->first<<" = "<<it->second<<")";
			lm << "(score = " << p->getScore() << ", contribution = " << contrib << ")";
		}
	}
	lastMessage = lm.str();
}

/** Return all the plans, ordered using the cost. The first is the best plan.
 * @return The list of plans
 */
vector<Plan*> PlanStockManager::getPlans() const {
	multimap<float, Plan*>::const_iterator it;
	vector<Plan*> vec;
	if(plans.size() > 0)
		for(it = plans.begin() ; it != plans.end() ; ++it) vec.push_back(it->second);
	return vec;
}

/** Return the current best plan.
 * @return The first member of plans
 */
Plan* PlanStockManager::getBestPlan() const {
	Plan* p = NULL;
	if(!plans.empty()) p = plans.begin()->second;
	return p;
}

/** Remove the first plan and return the next one.
 * @return The new best plan after removing the first one
 */
Plan* PlanStockManager::getNextPlan() {
	if(!plans.empty()) {
		plans.erase(plans.begin());
	}
		return this->getBestPlan();
}

/** Find a plan using its ID.
 * @param id Plan to find
 * @return The plan with the ID or NULL
 */
Plan* PlanStockManager::getPlanByID(unsigned long int id) const {
	multimap<float, Plan*>::const_iterator it;
	if(plans.size() > 0)
		for(it = plans.begin() ; it != plans.end() ; ++it) {
			if((it->second)->getID() == id) return it->second;
		}
	return NULL;
}

/** Return all the plan with the best score.
 * (There can be several plan with the lowest plan cost, the best score.)
 * @return All members of plans with the least score
 */
vector<Plan*> PlanStockManager::getBestPlans() const {
	multimap<float, Plan*>::const_iterator it;
	float min;
	vector<Plan*> vec;
	if(plans.size() > 0) {
		min = plans.begin()->first;
		for(it = plans.begin() ; it != plans.end() ; ++it) {
			if(it->first > min) break;
			vec.push_back(it->second);
		}
	}
	return vec;
}

/** Empty the plan stock.
 * Delete all members of plans and clear it.
 */
void PlanStockManager::clear() {
	for(multimap<float, Plan*>::iterator it=plans.begin() ; it!=plans.end() ; ++it)
		delete it->second;
	plans.clear();
	NumberOfPlans = 0;
	solutionExists = false;
}

/** Destructor: delete all members of plans and clear it
 */
PlanStockManager::~PlanStockManager() {
	for(multimap<float, Plan*>::iterator it=plans.begin() ; it!=plans.end() ; ++it)
		delete it->second;
}
