#include<iostream>

#include "WorldBaseAgent.hh"
#include "tools.hh"

using namespace std;

/** Add an Entity to the database.
 * If there already is an Entity with the same name it gives a warning and does not replace the previous.
 * @param e Entity to add to the database
 */
void WorldBaseAgent::add(Entity const& e) {
	if(database.find(e.getName()) == database.end()) database[e.getName()] = new Entity(e);
	else cerr << "WARNING: trying to add entity \"" << e.getName() << "\" whereas it already exists" << endl;
}

/** Clear all dynamic vectors of all the entities in the database.
 * It means that all dynamic attributes (of every entity) are set to their default value. (See AttributeSet::delallDynamicVector())
 */
void WorldBaseAgent::clearAllDynamicVectors() {
	for(map<string, Entity*>::iterator it=database.begin() ; it != database.end() ; ++it)
		(it->second)->delallDynamicVector();
}

/** Check if an Entity exists in the agent's database.
 * @param sName Name of the entity to check
 * @return True if the entity exists
 */
bool WorldBaseAgent::containEntity(std::string const& sName) const {
	return database.find(sName) != database.end();
}

/** Check if an Entity (with a specific type) exists in the agent's database.
 * @param sName Name of the entity to check
 * @param sType The type the entity need to be
 * @return True if the entity exists and is of the given type
 */
bool WorldBaseAgent::containEntity(std::string const& sName, std::string const& sType) const {
	map<string, Entity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType)
		return true;

	return false;
}

/** Find an Entity by its name in the agent's database.
 * @param sName Name of the entity to find
 * @return Pointer to the entity
 */
Entity* WorldBaseAgent::find(std::string const& sName) {
	std::map<std::string, Entity*>::iterator it = database.find(sName);
	if(it != database.end()) return it->second;
	return NULL;
}

/** Find an Entity by its name in the agent's database.
 * @param sName Name of the entity to find
 * @return Const pointer to the entity
 */
Entity const* WorldBaseAgent::find(std::string const& sName) const {
	std::map<std::string, Entity*>::const_iterator it = database.find(sName);
	if(it != database.end()) return it->second;
	return NULL;
}

/** Find an Entity by its name and type in the agent's database.
 * @param sName Name of the entity to find
 * @param sType Type of the entity
 * @return Pointer to the entity
 */
Entity* WorldBaseAgent::find(std::string const& sName, std::string const& sType) {
	map<string, Entity*>::iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType)
		return it->second;
	return NULL;
}

/** Find an Entity by its name and type in the agent's database.
 * @param sName Name of the entity to find
 * @param sType Type of the entity
 * @return Const pointer to the entity
 */
Entity const* WorldBaseAgent::find(std::string const& sName, std::string const& sType) const {
	map<string, Entity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType)
		return it->second;
	return NULL;
}

/** Find a set of all entities matching the given type.
 * @param sType The type the entities need to match
 * @return Set of entities matching the given type
 */
vector<Entity*> WorldBaseAgent::findByType(string const& sType) const {
	vector<Entity*> vec;
	for(map<string,Entity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		if(it->second->getType() == sType)
			vec.push_back(it->second);
	return vec;
}

/** Find a set of all entities that have a given attribute.
 * @param attrName The attribute the entities need to have
 * @return Set of entities having the given attribute
 */
vector<Entity*> WorldBaseAgent::findByAttribute(string const& attrName) const {
	vector<Entity*> vec;
	for(map<string,Entity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it) {
		if(it->second->hasAttributeOfName(attrName)) {
			vec.push_back(it->second);
		}
	}
	return vec;
}

/** Create a string representing the WorldBaseAgent.
 * @return A string representing the agent
 */
string WorldBaseAgent::toString() const {
	string s = "***** Start WorldBaseAgent Printing *****\n";
	for(map<string,Entity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		s += it->second->toString() + "\n";
	s += "***** End WorldBaseAgent Printing *****\n";
	return s;
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c Entity to return
 * @return Reference to the entity, or to an empty entity
 */
Entity& WorldBaseAgent::operator[](const char* c) {
	std::map<std::string, Entity*>::iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBaseAgent1: " << c;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s Entity to return
 * @return Reference to the entity, or to an empty entity
 */
Entity& WorldBaseAgent::operator[](std::string const& s) {
	std::map<std::string, Entity*>::iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBaseAgent2: " << s;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c Entity to return
 * @return Const reference to the entity, or to an empty entity
 */
Entity const& WorldBaseAgent::operator[](const char* c) const {
	std::map<std::string, Entity*>::const_iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBaseAgent3: " << c;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s Entity to return
 * @return Const reference to the entity, or to an empty entity
 */
Entity const& WorldBaseAgent::operator[](std::string const& s) const {
	std::map<std::string, Entity*>::const_iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBaseAgent4: " << s;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Destructor.
 * Delete all Entites in the database.
 */
WorldBaseAgent::~WorldBaseAgent() {
	for(map<string,Entity*>::iterator it=database.begin() ; it!=database.end() ; ++it)
		delete it->second;
}
