#include "WorldBaseUpdater.hh"

#include <boost/regex.hpp>

#include "RequestLexer.hh"
#include "tools.hh"
using namespace HATP_TOOLS;
#include "VerboseLevel.hh"

/** Ask for the value of a given entity's attribute for a given agent.
 * Only responsible of the communication, the parsing is done with update[Whole][BM]WorldBaseFromOntology functions.
 * @param agent Agent's database to prompt (can be "" if not using Belief Management).
 * @param object The entity.
 * @param property The entity's attribute.
 */
std::vector<std::string> WorldBaseUpdater::getUpdatedProperty(const std::string& agent, const std::string& object, const std::string& property){
	std::vector<std::string> results;
	std::ostringstream msg, request, reply;
	request << "{\"ONTO-REQ\": { ";
	if(agent == "") request << "\"AGENT\": null, ";
	else request << "\"AGENT\": \"" << agent << "\", ";
	request << "\"OBJECT\": \"" << object << "\", ";
	request << "\"PROPERTY\": \"" << property << "\"";
	reply << request.str() << ", \"VALUE\": ["; //The reply is formed as the request with a new field for the value
	request << "}}";
	mp.sendMessage(ontology_client, request.str());
	boost::smatch what; 
	std::pair<std::string, std::string> receivedMsg = mp.getBlockingMessage();
	std::string answer(receivedMsg.second);
	removeFormatting(answer);
	if(regex_match(answer, what, ONTOResult)) {
		std::string reqres = what[1];
		str_replace("\"", "", reqres);
		char *token, *cpar;
		cpar = new char[reqres.size()+1];
		strcpy (cpar, reqres.c_str());
		token = strtok(cpar,",");
		while (token != NULL) {
			results.push_back(token);
			token = strtok(NULL, ",");
		}
		msg << reply.str() << what[1] << "] }}";
		VERBOSE_2 << reply.str() << what[1] << "] }}" << std::endl;
		delete token;
		delete cpar;
	}
	else {
		msg << "format error of the Ontology answer";
		VERBOSE_2 << "format error of the Ontology answer" << std::endl;
	}
	mp.sendMessage(console_name, msg.str()); //Send the reply
	return results;
}

/** Update the whole WorldBase at once.
 * This update is done in the context of belief management. So the database will be requested for each agent.
 * @param Reference to the WorldBase.
 * @return True if the update worked correctly, false otherwise.
 * The format of the message allows to update only part of the database.
 * This function do it *all at once*, but a new function can easily be created to offer the other behavior.
 */
bool WorldBaseUpdater::updateWholeBMWorldBaseFromOntology(WorldBase& base) {
	bool updateOk=true;
	const std::map<std::string, Entity*>& entities = base.getDatabase();
	const std::vector<Entity*> agents = base.findByType("Agent");
	std::map<std::string, Entity*>::const_iterator itEntities;
	std::vector<Entity*>::const_iterator itAgents;
	std::vector<std::string> newValues ;
	
	VERBOSE_3 << "Updates the fact databse using ontology (in Belief Management context)" << std::endl;
	
	for(itEntities = entities.begin() ; itEntities != entities.end() ; ++itEntities) {
		Entity* current=itEntities->second;
		for(itAgents = agents.begin(); itAgents != agents.end(); ++itAgents) {
			std::string aname = (*itAgents)->getName();
			Entity* currentBM = base.findForAgent(aname, current->getName());
			if(currentBM == NULL) { // due to optim done in the domain parser
				base.initForAgent(aname, current->getName());
				currentBM = base.findForAgent(aname, current->getName());
			}
		}
	}

	for(itAgents = agents.begin(); itAgents != agents.end(); ++itAgents) {
		size_t numberOfProperties=0;

		//Send the request for all datas for this agent
		std::string aname=(*itAgents)->getName();
		std::ostringstream request;
		request << "{\"ONTO-REQ-ALL\": { \"AGENT\": \"" << aname << "\", \"OBJECTS\": [ ";
		
		bool firstEntity=true;
		for(itEntities = entities.begin() ; itEntities != entities.end() ; ++itEntities) {
			if(firstEntity) firstEntity = false;
			else request << ", ";
			
			//The hasKnownLocation is related to the belief management. From this property is infered the knowledge for the agent of the object (known or unknown)
			request << "{\"NAME\": \"" << itEntities->second->getName() << "\", \"PROPERTIES\": [ \"hasKnownLocation\"";
			++numberOfProperties;
			
			Entity* currentBM = base.findForAgent(aname, itEntities->second->getName());
			std::list<std::string> attributes(currentBM->getAllAttributeNames());
			
			for(std::list<std::string>::iterator itName = attributes.begin() ; itName != attributes.end(); itName++) {
				if(!currentBM->isDynamic(*itName)) {
					continue;
				}
				
				request << ", \"" << (*itName) << "\"";
				++numberOfProperties;
			}
			
			//End of the scope of the entity
			request << "] }";
		}
		
		request << "] }}";
		bool result=mp.sendMessage(ontology_client, request.str());
		if(!result){
			std::cerr << "Error: Impossible to reach the database " << ontology_client << std::endl;
			return false;
		}
		
		std::pair<std::string, std::string> receivedMsg = mp.getBlockingMessage();
		boost::smatch what;
		std::string answer(receivedMsg.second);
		removeFormatting(answer);
		std::ostringstream msg;
		std::list<std::string> valueFields;
		
		VERBOSE_3<<"\nAnswer to request database:\n"<<answer<<"\n\n";
	
		if(regex_match(answer, what, ONTOALLResult)) {
			std::string allvalues(what[1]);
			{
				boost::sregex_token_iterator valueScopeIterator(allvalues.begin(), allvalues.end(), ONTOValueScope), endScopeIterator;
		
				for(;valueScopeIterator!=endScopeIterator; ++valueScopeIterator){
					std::string temp(*valueScopeIterator);
					size_t posFirst, posLast;
					posFirst=temp.find("[");
					posLast=temp.rfind("]");
					if(posFirst!=std::string::npos && posLast!=std::string::npos){
						if(posFirst+1!=posLast){
							valueFields.push_back(temp.substr(posFirst+1, posLast-posFirst-1));
						}else{
							valueFields.push_back("");
						}
					}
				}
		
				if(valueFields.size()!=numberOfProperties){
					msg << "format error of the Ontology answer, wrong number of value fields. Wanted:" << valueFields.size() << " Found:" << numberOfProperties;
					VERBOSE_2 << msg.str() << std::endl;
					mp.sendMessage(console_name, msg.str());
					return false;
				}
			}
		
			//Go through all attributes in the same order there were used to build the request
			std::string msgHeader("{\"ONTO-REQ\": { \"AGENT\": \""+aname+"\", ");
			for(itEntities = entities.begin() ; itEntities != entities.end() ; ++itEntities) {
				Entity* currentBM = base.findForAgent(aname, itEntities->second->getName());

				//The very first element of the valueFields list contain the "hasKnownLocation" fact, its default value is true
				std::string value="";
				if(valueFields.front().empty() || valueFields.front()=="\"true\""){
					currentBM->setKnown(true);
					value="\"known\"";
				}else{ //false
					currentBM->setUnknown(true);
					value="\"unknown\"";
				}
				mp.sendMessage(console_name, msgHeader+"\"OBJECT\": \""+currentBM->getName()+"\", \"PROPERTY\": \"hasKnownLocation\", \"VALUE\": ["+value+"] }}");
				if(valueFields.size()!=0) valueFields.pop_front();

				std::list<std::string> attributes(currentBM->getAllAttributeNames());
				
				for(std::list<std::string>::iterator itName = attributes.begin(); itName != attributes.end(); ++itName){
					if(!currentBM->isDynamic(*itName)) {
						continue;
					}
					std::ostringstream reply;
					reply << msgHeader << "\"OBJECT\": \"" << currentBM->getName() << "\", ";
					reply << "\"PROPERTY\": \"" << *itName << "\", \"VALUE\": [" << valueFields.front() << "] }}";
				
					//Interpret the values
					std::string scope(valueFields.front());
					boost::sregex_token_iterator valueIterator(scope.begin(), scope.end(), ONTOValue), endValueIterator;
					
					if(currentBM->isAtom(*itName)){
						if(valueIterator!=endValueIterator && *valueIterator=="unknown"){
							currentBM->setAttrUnknown(*itName, true);
						}else if(valueIterator!=endValueIterator && *valueIterator=="known"){
							currentBM->setAttrKnown(*itName, true);
						}else if(valueIterator!=endValueIterator && currentBM->getNumberAttrs().hasAttributeOfName(*itName)){
							number nval; std::istringstream iss(*valueIterator); iss >> nval;
							currentBM->get<number>(*itName) = nval;
						}else if(currentBM->getStringAttrs().hasAttributeOfName(*itName)){
							if(valueIterator==endValueIterator){
								currentBM->get<std::string>(*itName) = "";
							}else{
								currentBM->get<std::string>(*itName) = *valueIterator;
							}
						}else if(valueIterator!=endValueIterator && currentBM->getBoolAttrs().hasAttributeOfName(*itName)){
							(*valueIterator == "true")?(currentBM->get<bool>(*itName) = true):
								(currentBM->get<bool>(*itName) = false);
						}else {//(currentBM->getEntityAttrs().hasAttributeOfName(*itName))
							if(valueIterator==endValueIterator || *valueIterator=="NULL"){
								currentBM->get<Entity*>(*itName) = NULL;
							}else{
								currentBM->get<Entity*>(*itName) = base.find(*valueIterator);
								if(currentBM->get<Entity*>(*itName)==NULL){
									std::cerr<<"Error the Entity \""<<*valueIterator<<"\" does not exist but is used in the database to initialise "<<currentBM->getName()<<"."<<*itName<<std::endl;
									return false;
								}
							}
						}
						if(valueIterator!=endValueIterator && ++valueIterator!=endValueIterator){
							std::cerr<<"Error while retrieving database, wrong type for "<<currentBM->getName()<<"."<<*itName<<std::endl;
							return false;
						}
					}else{ //isSet
						if(currentBM->getNumberAttrs().hasAttributeOfName(*itName)){
							currentBM->get< std::set<number> >(*itName).clear();
						}else if(currentBM->getStringAttrs().hasAttributeOfName(*itName)){
							currentBM->get< std::set<std::string> >(*itName).clear();
						}else if(currentBM->getBoolAttrs().hasAttributeOfName(*itName)){
							currentBM->get< std::set<bool> >(*itName).clear();
						}else {//(currentBM->getEntityAttrs().hasAttributeOfName(*itName))
							currentBM->get< std::set<Entity*> >(*itName).clear();
						}
					
						if(valueIterator!=endValueIterator){//The variable is a set but the value field is empty
							for(;valueIterator!=endValueIterator; ++valueIterator){
								if(currentBM->getNumberAttrs().hasAttributeOfName(*itName)){
									number nval; std::istringstream iss(*valueIterator); iss >> nval;
									currentBM->get< std::set<number> >(*itName) <<= nval;
								}else if(currentBM->getStringAttrs().hasAttributeOfName(*itName)){
									currentBM->get< std::set<std::string> >(*itName) <<= std::string(*valueIterator);
								}else if(currentBM->getBoolAttrs().hasAttributeOfName(*itName)){
									(*valueIterator == "true")?(currentBM->get< std::set<bool> >(*itName) <<= true):
										(currentBM->get< std::set<bool> >(*itName) <<= false);
								}else {//(currentBM->getEntityAttrs().hasAttributeOfName(*itName))
									if(base.findForAgent (aname, *valueIterator)==NULL){
										std::cerr<<"Error the Entity \""<<*valueIterator<<"\" does not exist but is used in the database to initialise "<<currentBM->getName()<<"."<<*itName<<std::endl;
										return false;
									}
									currentBM->get< std::set<Entity*> >(*itName) <<= base.findForAgent (aname, *valueIterator);
								}
							}
						}
					}
				
					mp.sendMessage(console_name, reply.str());
					valueFields.pop_front();
				}
			}
		} else {
			msg << "format error of the Ontology answer";
			VERBOSE_2 << "format error of the Ontology answer" << std::endl;
			updateOk=false;
		}
		
		mp.sendMessage(console_name, msg.str()); //Send the reply
	}
		
	return updateOk;
}

/** Update the whole WorldBase at once.
 * @param Reference to the WorldBase.
 * @return True if the update worked correctly, false otherwise.
 * The format of the message allows to update only part of the database.
 * This function do it *all at once*, but a new function can easily be created to offer the other behavior.
 */
bool WorldBaseUpdater::updateWholeWorldBaseFromOntology(WorldBase& base) {
	bool updateOk=true;
	size_t numberOfProperties=0;
	const std::map<std::string, Entity*>& entities = base.getDatabase();
	std::map<std::string, Entity*>::const_iterator itEntities;
	std::vector<std::string> newValues ;
	
	VERBOSE_3 << "Updates the fact databse using ontology" << std::endl;
	
	//Send the request for all datas for this agent
	std::ostringstream request;
	request << "{\"ONTO-REQ-ALL\": { \"AGENT\": null, \"OBJECTS\": [ ";
	
	bool firstEntity=true;
	for(itEntities = entities.begin() ; itEntities != entities.end() ; ++itEntities) {
		Entity* current = itEntities->second;
		if(firstEntity) firstEntity = false;
		else request << ", ";
		
		request << "{\"NAME\": \"" << current->getName() << "\", \"PROPERTIES\": [ ";
		
		std::list<std::string> attributes(current->getAllAttributeNames());
		
		bool firstProperty=true;
		for(std::list<std::string>::iterator itName = attributes.begin() ; itName != attributes.end(); itName++) {
			if(!current->isDynamic(*itName)) {
				continue;
			}
			
			//As the property is dynamic, its value is requested
			if(firstProperty) firstProperty = false;
			else request << ", ";
			
			request << "\"" << (*itName) << "\"";
			++numberOfProperties;
		}
		
		//End of the scope of the entity
		request << "] }";
	}
	
	request << "] }}";
	bool result=mp.sendMessage(ontology_client, request.str());
	if(!result){
		std::cerr << "Error: Impossible to reach the database " << ontology_client << std::endl;
		return false;
	}
	
	std::pair<std::string, std::string> receivedMsg = mp.getBlockingMessage();
	boost::smatch what;
	std::string answer(receivedMsg.second);
	removeFormatting(answer);
	std::ostringstream msg;
	std::list<std::string> valueFields;
	
	if(regex_match(answer, what, ONTOALLResult)) {
		std::string allvalues(what[1]);
		{
			boost::sregex_token_iterator valueScopeIterator(allvalues.begin(), allvalues.end(), ONTOValueScope), endScopeIterator;
		
			for(;valueScopeIterator!=endScopeIterator; ++valueScopeIterator){
				std::string temp(*valueScopeIterator);
				size_t posFirst, posLast;
				posFirst=temp.find("[");
				posLast=temp.rfind("]");
				if(posFirst!=std::string::npos && posLast!=std::string::npos){
					if(posFirst+1!=posLast){
						valueFields.push_back(temp.substr(posFirst+1, posLast-posFirst-1));
					}else{
						valueFields.push_back("");
					}
				}
			}
		
			if(valueFields.size()!=numberOfProperties){
				msg << "format error of the Ontology answer, wrong number of value fields. Wanted:" << valueFields.size() << " Found:" << numberOfProperties;
				VERBOSE_2 << msg.str() << std::endl;
				mp.sendMessage(console_name, msg.str());
				return false;
			}
		}
		
		//Go through all attributes in the same order there were used to build the request
		std::string msgHeader("{\"ONTO-REQ\": { \"AGENT\": null, ");
		for(itEntities = entities.begin() ; itEntities != entities.end() ; ++itEntities) {
			Entity* current = itEntities->second;
			std::list<std::string> attributes(current->getAllAttributeNames());
			
			for(std::list<std::string>::iterator itName = attributes.begin(); itName != attributes.end(); ++itName){
				if(!current->isDynamic(*itName)) {
					continue;
				}
				std::ostringstream reply;
				reply << msgHeader << "\"OBJECT\": \"" << current->getName() << "\", ";
				reply << "\"PROPERTY\": \"" << *itName << "\", \"VALUE\": [" << valueFields.front() << "] }}";
				
				//Interpret the values
				std::string scope(valueFields.front());
				boost::sregex_token_iterator valueIterator(scope.begin(), scope.end(), ONTOValue), endValueIterator;
				
				if(current->isAtom(*itName)){
					if(valueIterator!=endValueIterator && current->getNumberAttrs().hasAttributeOfName(*itName)){
						number nval; std::istringstream iss(*valueIterator); iss >> nval;
						current->get<number>(*itName) = nval;
					}else if(current->getStringAttrs().hasAttributeOfName(*itName)){
						if(valueIterator==endValueIterator){
							current->get<std::string>(*itName) = "";
						}else{
							current->get<std::string>(*itName) = *valueIterator;
						}
					}else if(valueIterator!=endValueIterator && current->getBoolAttrs().hasAttributeOfName(*itName)){
						(*valueIterator == "true")?(current->get<bool>(*itName) = true):
							(current->get<bool>(*itName) = false);
					}else {//(current->getEntityAttrs().hasAttributeOfName(*itName))
						if(valueIterator==endValueIterator || *valueIterator=="NULL"){
							current->get<Entity*>(*itName) = NULL;
						}else{
							current->get<Entity*>(*itName) = base.find(*valueIterator);
							if(current->get<Entity*>(*itName)==NULL){
								std::cerr<<"Error the Entity \""<<*valueIterator<<"\" does not exist but is used in the database to initialise "<<current->getName()<<"."<<*itName<<std::endl;
								return false;
							}
						}
					}
					if(valueIterator!=endValueIterator && ++valueIterator!=endValueIterator){
						std::cerr<<"Error while retrieving database, wrong type for "<<current->getName()<<"."<<*itName<<std::endl;
					}
				}else{ //isSet
					if(current->getNumberAttrs().hasAttributeOfName(*itName)){
						current->get< std::set<number> >(*itName).clear();
					}else if(current->getStringAttrs().hasAttributeOfName(*itName)){
						current->get< std::set<std::string> >(*itName).clear();
					}else if(current->getBoolAttrs().hasAttributeOfName(*itName)){
						current->get< std::set<bool> >(*itName).clear();
					}else {//(current->getEntityAttrs().hasAttributeOfName(*itName))
						current->get< std::set<Entity*> >(*itName).clear();
					}
					
					if(valueIterator!=endValueIterator){//The variable is a set but the value field is empty
						for(;valueIterator!=endValueIterator; ++valueIterator){
							if(current->getNumberAttrs().hasAttributeOfName(*itName)){
								number nval; std::istringstream iss(*valueIterator); iss >> nval;
								current->get< std::set<number> >(*itName) <<= nval;
							}else if(current->getStringAttrs().hasAttributeOfName(*itName)){
								current->get< std::set<std::string> >(*itName) <<= std::string(*valueIterator);
							}else if(current->getBoolAttrs().hasAttributeOfName(*itName)){
								(*valueIterator == "true")?(current->get< std::set<bool> >(*itName) <<= true):
									(current->get< std::set<bool> >(*itName) <<= false);
							}else {//(current->getEntityAttrs().hasAttributeOfName(*itName))
//TODO								if(value=="true" || value=="false") { //Hack HATP/ORO : agent-centered onto
//									std::vector<Entity*> agents = base.findByType("Agent");
//									for(std::vector<Entity*>::iterator it=agents.begin(); it!=agent.end(); ++it){
//										std::string agentName = (*it)->getName();
//										std::vector<std::string> agentValues = getUpdatedProperty(mp, ontology_client, agentName, current->getName(), *itName);
//										if(!agentValues.empty() && agentValues[0] == "true") {
//											current->get< std::set<Entity*> >(*itName) <<= agents[i];
//										}
//									}
//								}
//								else {
									if(base.find(*valueIterator)==NULL){
										std::cerr<<"Error the Entity \""<<*valueIterator<<"\" does not exist but is used in the database to initialise "<<current->getName()<<"."<<*itName<<std::endl;
										return false;
									}
									current->get< std::set<Entity*> >(*itName) <<= base.find(*valueIterator);
//								}
							}
						}
					}
				}
				
				mp.sendMessage(console_name, reply.str());
				valueFields.pop_front();
			}
		}
	} else {
		msg << "format error of the Ontology answer";
		VERBOSE_2 << "format error of the Ontology answer" << std::endl;
		updateOk=false;
	}
		
	mp.sendMessage(console_name, msg.str()); //Send the reply
	return updateOk;
}

/**Update the database asking one entity's atrtibute after another.
 * In order to ensure retro-compatibility with CHRIS experiment.
 * We separate the world update with Belief Management from the classical update.
 * @return True if the update worked correctly, false otherwise.
 * DEPRECATED.
 */
bool WorldBaseUpdater::updateBMWorldBaseFromOntology(WorldBase& base) {
	bool updateOK = true;
	std::map<std::string, Entity*> entities = base.getDatabase();
	std::vector<Entity*> agents = base.findByType("Agent");
	std::map<std::string, Entity*>::iterator it;
	std::vector<std::string> newValues ;

	for(it=entities.begin() ; it!=entities.end() ; it++) {
		Entity* current = it->second;
		for(unsigned int i = 0 ; i < agents.size() ; i ++) {
			std::string aname = agents[i]->getName();
			Entity* currentBM = base.findForAgent(aname, current->getName());
			if(currentBM == NULL) { // due to optim done in the domain parser
				base.initForAgent(aname, current->getName());
				currentBM = base.findForAgent(aname, current->getName());
			}
			// hack matthieu for Belief Management via hasKnownLocation
			newValues = getUpdatedProperty(aname, currentBM->getName(), "hasKnownLocation");
			if(!newValues.empty()) {
				if(newValues[0] == "true") currentBM->setKnown(true);
				if(newValues[0] == "false") currentBM->setUnknown(true);
			}
		}
	}

	for(it=entities.begin() ; it!=entities.end() ; it++) {
		Entity* current = it->second;
		for(unsigned int i = 0 ; i < agents.size() ; i ++) {
			std::string aname = agents[i]->getName();
			Entity* currentBM = base.findForAgent(aname, current->getName());
			// attributes in agent model
			std::list<std::string> attributes = currentBM->getAllAttributeNames();
			for(std::list<std::string>::iterator itName = attributes.begin() ; itName != attributes.end(); itName++) {
				if(currentBM->isDynamic(*itName)) {
					// request to ORO
					newValues = getUpdatedProperty(aname, currentBM->getName(), *itName);
					// end request
					bool notBMVal = true;
					if(!newValues.empty()) {
						std::string testValue(newValues[0]);
						if(testValue == "unknown") {currentBM->setAttrUnknown(*itName, true); notBMVal = false;}
						if(testValue == "known") {currentBM->setAttrKnown(*itName, true);notBMVal = false;}
						if(currentBM->isAtom(*itName) && notBMVal) { // atom management 
							std::string value(newValues[0]);
							if(currentBM->getNumberAttrs().hasAttributeOfName(*itName)) { // is number
								number nval; std::istringstream iss(value); iss >> nval;
								currentBM->get<number>(*itName) = nval; }
							if(currentBM->getStringAttrs().hasAttributeOfName(*itName)) { // is string
								currentBM->get<std::string>(*itName) = value; }
							if(currentBM->getBoolAttrs().hasAttributeOfName(*itName)) { // is boolean
								(value == "true")?(currentBM->get<bool>(*itName) = true):
									(currentBM->get<bool>(*itName) = false); }
							if(currentBM->getEntityAttrs().hasAttributeOfName(*itName)) { // is entity
								currentBM->get<Entity*>(*itName) = base.findForAgent(aname, value); }
						}
						if(currentBM->isSet(*itName) && notBMVal) { // value sets management
							if(currentBM->getNumberAttrs().hasAttributeOfName(*itName)) (currentBM->get< std::set<number> >(*itName)).clear();
							if(currentBM->getStringAttrs().hasAttributeOfName(*itName)) (currentBM->get< std::set<std::string> >(*itName)).clear();
							if(currentBM->getBoolAttrs().hasAttributeOfName(*itName)) (currentBM->get< std::set<bool> >(*itName)).clear();
							if(currentBM->getEntityAttrs().hasAttributeOfName(*itName)) (currentBM->get< std::set<Entity*> >(*itName)).clear();	
							for(std::vector<std::string>::iterator itVal = newValues.begin() ; itVal != newValues.end() ; itVal++) {
								std::string value = *itVal;
								if(currentBM->getNumberAttrs().hasAttributeOfName(*itName)) { // is number
									number nval; std::istringstream iss(value); iss >> nval;
									currentBM->get< std::set<number> >(*itName) <<= nval; }
								if(currentBM->getStringAttrs().hasAttributeOfName(*itName)) { // is string
									currentBM->get< std::set<std::string> >(*itName) <<= value; }
								if(currentBM->getBoolAttrs().hasAttributeOfName(*itName)) { // is boolean
									(value == "true")?(current->get< std::set<bool> >(*itName) <<= true):
										(currentBM->get< std::set<bool> >(*itName) <<= false); }
								if(currentBM->getEntityAttrs().hasAttributeOfName(*itName)) { // is entity
									Entity* valtemp = base.findForAgent(aname, value);
									if(valtemp != NULL) currentBM->get< std::set<Entity*> >(*itName) <<= base.findForAgent(aname, value); 
								}
							} // for Val
						} // isSet
					} // !empty
				} // isDynamic
			} // for attr
		} // for agents
	} // for entities
	return updateOK;
}

/**Update the database asking one entity's atrtibute after another.
 * In order to ensure retro-compatibility with CHRIS experiment.
 * We separate the world update with Belief Management from the classical update.
 * @return True if the update worked correctly, false otherwise.
 * DEPRECATED.
 */
bool WorldBaseUpdater::updateWorldBaseFromOntology(WorldBase& base) {
	bool updateOK = true;
	std::map<std::string, Entity*> entities = base.getDatabase();
	std::map<std::string, Entity*>::iterator it;
	std::vector<std::string> newValues ;
	for(it=entities.begin() ; it!=entities.end() ; it++) {
		Entity* current = it->second;
		std::list<std::string> attributes = current->getAllAttributeNames();
		for(std::list<std::string>::iterator itName = attributes.begin() ; itName != attributes.end(); itName++) {
			if(current->isDynamic(*itName)) {
				newValues = getUpdatedProperty("", current->getName(), *itName);
				if(!newValues.empty()) {
					if(current->isAtom(*itName)) { // atom management
						std::string value(newValues[0]);
						if(current->getNumberAttrs().hasAttributeOfName(*itName)) { // is number
							number nval; std::istringstream iss(value); iss >> nval;
							current->get<number>(*itName) = nval; }
						if(current->getStringAttrs().hasAttributeOfName(*itName)) { // is string
							current->get<std::string>(*itName) = value; }
						if(current->getBoolAttrs().hasAttributeOfName(*itName)) { // is boolean
							(value == "true")?(current->get<bool>(*itName) = true):
								(current->get<bool>(*itName) = false); }
						if(current->getEntityAttrs().hasAttributeOfName(*itName)) { // is entity
							current->get<Entity*>(*itName) = base.find(value); }
					}
					if(current->isSet(*itName)) { // value sets management
						if(current->getNumberAttrs().hasAttributeOfName(*itName)) (current->get< std::set<number> >(*itName)).clear();
						if(current->getStringAttrs().hasAttributeOfName(*itName)) (current->get< std::set<std::string> >(*itName)).clear();
						if(current->getBoolAttrs().hasAttributeOfName(*itName)) (current->get< std::set<bool> >(*itName)).clear();
						if(current->getEntityAttrs().hasAttributeOfName(*itName)) (current->get< std::set<Entity*> >(*itName)).clear();	
						for(std::vector<std::string>::iterator itVal = newValues.begin() ; itVal != newValues.end() ; itVal++) {
							std::string value = *itVal;
							if(current->getNumberAttrs().hasAttributeOfName(*itName)) { // is number
								number nval; std::istringstream iss(value); iss >> nval;
								current->get< std::set<number> >(*itName) <<= nval; }
							if(current->getStringAttrs().hasAttributeOfName(*itName)) { // is string
								current->get< std::set<std::string> >(*itName) <<= value; }
							if(current->getBoolAttrs().hasAttributeOfName(*itName)) { // is boolean
								(value == "true")?(current->get< std::set<bool> >(*itName) <<= true):
									(current->get< std::set<bool> >(*itName) <<= false); }
							if(current->getEntityAttrs().hasAttributeOfName(*itName)) { // is entity
								if(value=="true" || value=="false") { //ici hack HATP/ORO : agent-centered onto
									std::vector<Entity*> agents = base.findByType("Agent");
									for(unsigned int i = 0 ; i < agents.size() ; i++) {
										std::string agentName = agents[i]->getName();
										std::vector<std::string> agentValues = getUpdatedProperty(agentName, current->getName(), *itName);
										if(!agentValues.empty() && agentValues[0] == "true") {
											current->get< std::set<Entity*> >(*itName) <<= agents[i];
										}
									}
								}
								else { current->get< std::set<Entity*> >(*itName) <<= base.find(value); }
							}
						} // for Val
					} // isSet
				} // !empty
			} // isDynamic
		} // for attr
	} // for entities
	return updateOK;
}
